/**
 * Main.cpp
**/

#include <stdio.h>
#include <glut.h>
#include <gl\gl.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>

#include "world.h"
#include "renderer.h"

// Yucky globals
World *world;
Renderer *renderer;
int main_window;

// Precalculate this stuff for faster lookup times
double cos_lookup[360];
double sin_lookup[360];

int USE_GAMEMODE = 0;


// Function defs
void MenuSelect( int value );
void InitWindow();
void InitStuff( int argc, char* argv[], int dbuffer );

/**
 * The display function. It is called whenever the window needs
 * redrawing (ie: overlapping window moves, resize, maximize)
**/
void display(void)
{
	renderer->RenderWorld();

	glutPostRedisplay();
}

/**
 * Timer function used to update the world at regular intervals,
 * independent of the FPS
**/
void Timer(int value)
{
	world->UpdateWorld();
	//printf("timer");
		// 30 fps game timer
  glutTimerFunc( 33, Timer, 1 );
}

/**
 * This function is called whenever the window is resized. 
 * Parameters are the new dimentions of the window
**/
void	resize(int x,int y)
{
	// Thanks to lesson #7
	glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
	glLoadIdentity();									// Reset The Projection Matrix

	// Calculate The Aspect Ratio Of The Window
	//gluPerspective(45.0f,(GLfloat)x/(GLfloat)y,0.1f,100.0f);
	//gluPerspective(90.0f,(GLfloat)x/(GLfloat)y,0.1f,100.0f);
	//gluPerspective(45.0f,(GLfloat)x/(GLfloat)y,0.1f,1000.0f);

	// View frustrum size depends on fog
	if( renderer->use_fog )
		gluPerspective(45.0f,(GLfloat)x/(GLfloat)y,0.1f,300.0f);
	else
		gluPerspective(45.0f,(GLfloat)x/(GLfloat)y,0.1f,1000.0f);
	
	glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	glLoadIdentity();									// Reset The Modelview Matrix

	glViewport(0,0,x,y);
}

/**
 * This function is called whenever the mouse is pressed or released
 * button is a number 0 to 2 designating the button
 * state is 1 for release 0 for press event
**/
void	mouse(int button,int state,int x,int y)
{
	// Temp
  //display();
}

void PassiveMouseMotion( int x, int y )
{
	// Maybe this should redraw the cube as it's being rotated?
}

void BoringKeyDown(unsigned char key, int x, int y)
{
	world->keys[key] = 1;

	// ESC key in game mode will take you back to window mode
	if( key == 27 && USE_GAMEMODE == 1 ) {
		USE_GAMEMODE = 0;
		glutLeaveGameMode();
		InitWindow();
		renderer->InitGLStuff();
	}

}

void SpecialKeyDown(int key, int x, int y)
{
	world->keys[key] = 1;

}

void SpecialKeyUp(int key, int x, int y)
{
	world->keys[key] = 0;
}


void MenuSelect( int value )
{
	switch( value ) {

	case 11:
		world->camera_mode = CAMERA_BIRD;
		break;

	case 12:
		world->camera_mode = CAMERA_SMART;
		break;

	case 20:
		// Enable flog
		renderer->use_fog = 1;
		glutChangeToMenuEntry( 3, "+ Fog", 21 );
		glClearColor(0.7f, 0.7f, 0.7f, 0.7f);		// Grey foggy background
		glDepthRange( 0.1f, 300.0f );
		break;

	case 21:
		// Disable
		renderer->use_fog = 0;
		glutChangeToMenuEntry( 3, "- Fog", 20 );
		glClearColor(0.2f, 0.2f, 1.0f, 0.7f);		// Blue sky background
		glDepthRange( 0.1f, 1000.0f );
		break;

	case 30:
		// Disable shadows
		renderer->use_shadows = 0;
		glutChangeToMenuEntry( 4, "- Shadows", 31 );
		break;

	case 31:
		// Enable shadows
		renderer->use_shadows = 1;
		glutChangeToMenuEntry( 4, "+ Shadows", 30 );
		break;
		break;

	case 40:
		// Disable FPS
		renderer->use_fps = 0;
		glutChangeToMenuEntry( 5, "- FPS", 41 );
		break;

	case 41:
		// Enable FPS
		renderer->use_fps = 1;
		glutChangeToMenuEntry( 5, "+ FPS", 40 );
		break;

	case 80:
		// Regenerate new, bumpy land.
		world->terrain->use_flatland = 1;
		world->terrain->Generate();
		renderer->RenderTerrainList();
		glutChangeToMenuEntry( 6, "- Bumps and Hills", 81 );
		break;

	case 81:
		// Regenerate new, flat land.
		world->terrain->use_flatland = 0;
		world->terrain->Generate();
		renderer->RenderTerrainList();
		glutChangeToMenuEntry( 6, "+ Bumps and Hills", 80 );
		break;

	case 100:
		glutDestroyWindow( main_window );
		USE_GAMEMODE = 1;
		InitWindow();
		renderer->InitGLStuff();
		break;

	case 101:
		// This will never happen... MUHAHAHAHAHAHHAHAH!
		glutLeaveGameMode();
		
		break;

	default:
		// Blah!
		break;

	}
}

void InitWindow()
{
	// Full screen?
	if( USE_GAMEMODE ) {
		glutGameModeString("640x480:32");
		glutEnterGameMode();
	}
	else
		main_window = glutCreateWindow("GLBoids");

  glutDisplayFunc(display);
  glutReshapeFunc(resize);
  glutMouseFunc(mouse);
	glutPassiveMotionFunc(PassiveMouseMotion);
	glutSpecialUpFunc( SpecialKeyUp );
	glutSpecialFunc( SpecialKeyDown );
	glutKeyboardFunc( BoringKeyDown );

	// 30 fps game timer
  glutTimerFunc( 33, Timer, 1 );

	if( !USE_GAMEMODE ) {
		int view_submenu = glutCreateMenu( MenuSelect );
		glutAddMenuEntry("Bird Control", 11);
		glutAddMenuEntry("Smart Follow", 12);
			
		glutCreateMenu( MenuSelect );
		glutAddSubMenu("View", view_submenu);
		glutAddMenuEntry("----------", -1);
		glutAddMenuEntry("+ Fog", 21);
		glutAddMenuEntry("+ Shadows", 30);
		glutAddMenuEntry("+ FPS", 40);
		glutAddMenuEntry("+ Bumps and Hills", 80);
		glutAddMenuEntry("----------", -1);
		glutAddMenuEntry("Full Screen Mode", 100);

		glutAttachMenu(GLUT_RIGHT_BUTTON);
	}
    
}

void CalculateLookupTables()
{
	for( int i = 0; i < 360; i++ ) {
		cos_lookup[i] = cos( i / (float)180 * 3.1415 );
		sin_lookup[i] = sin( i / (float)180 * 3.1415 );
	}
}

/**
 * @param dbuffer Should it be doublebuffered?
**/
void InitStuff( int argc, char* argv[], int dbuffer )
{
	// Randomize crap
	time_t timer;
	srand( time(&timer) );

	// Initialize GLUT
  glutInit(&argc, argv);
	glutInitWindowSize(640, 480);
	if( dbuffer )
		//glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE | GLUT_ACCUM);
		glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE );
	else
		glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH );

	InitWindow();

	CalculateLookupTables();
}

void PrintCrap(void)
{
	printf("---- GLBoids ---- \n");
	printf("Tom Bak\nThong Chau\n\n\n");
	printf("Use the arrow keys to control your boid.\n");
	printf("The right mouse button brings up a context menu.");
	printf("\n\n");
	printf("*** NOTE: 3D OpenGL Accelerated video card highly recomended!!!\n\n");
}

// Here's the main
int main(int argc, char* argv[])
{
	PrintCrap();

	// Need to declare this first to get use_doublebuffer.
	renderer = new Renderer();	

	// Init misc window crap
	InitStuff(argc, argv, renderer->use_doublebuffer );

	world = new World();
 
	renderer->SetWorld( world );
	renderer->InitGLStuff();

 	// Switch to main loop
  glutMainLoop();

	// How do we get here?
	if( USE_GAMEMODE )
		glutLeaveGameMode();

  delete world;

  return 0;        
}
