/**
 * Renderer.Cpp
**/

#include "renderer.h"
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <math.h>

Renderer::Renderer()
{
  use_fog = 1;
	use_textures = 1;
	use_lighting = 0;
	use_sky = 0;
	use_billboards = 1;
	use_fps = 1;
	use_doublebuffer = 1;
	use_shadows = 1;
	LANDCHUNK_WIDTH = 20.0;
}

Renderer::~Renderer()
{

}

/**
 * Draw a simple 2D sky
**/
void Renderer::RenderSky()
{
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
  glDepthMask( 0 );

	// Set blending parameters
	glColor4f(1.0, 1.0, 1.0, 0.5f);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glEnable(GL_BLEND);

	//glPixelZoom( 2, 1 );
	// This is really fucking slow!!!
  //glDrawPixels( tex_cloud->width, tex_cloud->height, GL_RGBA, GL_UNSIGNED_BYTE, tex_cloud->data );
  glDepthMask( 1 );

	glDisable(GL_BLEND);
  
}

/**
 * @return Random float from 0.0 - 1.0
**/
float Renderer::RandFloat()
{
  return (float)( rand() % 1000 ) / (float) 1000;
}


void Renderer::RenderTerrainList()
{

	// Good ol terrain
	glNewList(dl_terrain, GL_COMPILE);
	
	// Need full white light for some reason???
	glColor3f(1.0, 1.0, 1.0);

	if( use_textures ) {
	  glEnable(GL_TEXTURE_2D);
	  glBindTexture( GL_TEXTURE_2D, tex_land->id );
	}

	// X position for previous and current row
  float tposx0, tposx1;
	float tposy0, tposy1;
	
  Terrain *ter = world->terrain;

	// We actually handle two rows at a time (since a square is formed)
	for( int i = 0; i < TERRAIN_SIZE-1; i++ ) {
		// Start up this row
		tposx0 = 0.0f;
		tposx1 = 0.0f;
		tposy0 = 0.0f;
		tposy1 = 0.0f;

		// Calculate starting y positions for the two rows
		for( int k = 0; k < i; k++ ) {
			tposy0 += ter->tex_y[0][k];
	//		if( tposy0 > 1.0 )
		//		tposy0 -= 1.0;
		}
		tposy1 = tposy0 + ter->tex_y[0][i];
//		if( tposy1 > 1.0f )
//			tposy1 -= 1.0;
		
	
		// Each triangle strip is a row of land
		glBegin(GL_TRIANGLE_STRIP);

		// back left
		glTexCoord2f( tposx0, tposy0 );
		glVertex3f(0 * LANDCHUNK_WIDTH, 10 * ter->heights[0][i], i * LANDCHUNK_WIDTH);
			
		// forward left
		glTexCoord2f( tposx1, tposy1 );
		glVertex3f(0 * LANDCHUNK_WIDTH, 10 * ter->heights[0][i+1], (i+1) * LANDCHUNK_WIDTH);

		for( int j = 0; j < TERRAIN_SIZE-1; j++ ) {
			
			// back right
			//glTexCoord2f( tposx0 + world->terrain->tex_x[j][i], tposy );
			glTexCoord2f( tposx0 + world->terrain->tex_x[j][i], tposy0 + world->terrain->tex_y[j][i] );
			glVertex3f( (j+1) * LANDCHUNK_WIDTH, 10 * ter->heights[j+1][i], i * LANDCHUNK_WIDTH);

			// forward right
			glTexCoord2f( tposx1 + world->terrain->tex_x[j][i+1], tposy1 + ter->tex_y[j][i+1] );
			glVertex3f( (j+1) * LANDCHUNK_WIDTH, 10 * ter->heights[j+1][i+1], (i+1) * LANDCHUNK_WIDTH);

  		tposx0 += ter->tex_x[j][i];
			
//			if( tposx0 > 1.0 )
//				tposx0 -= 1.0;

			tposx1 += ter->tex_x[j][i+1];
//			if( tposx1 > 1.0 )
//				tposx1 -= 1.0;


			tposy0 += ter->tex_y[j][i];
//			if( tposy0 > 1.0 )
//				tposy0 -= 1.0;
			tposy1 += ter->tex_y[j][i+1];
//			if( tposy1 > 1.0 )
//				tposy1 -= 1.0;
				
		}

		glEnd();

	}

	glDisable(GL_TEXTURE_2D);

  glEndList();
}

void Renderer::RenderBoids()
{
	for( int i = 0; i < NUM_BOIDS; i++ ) {
		
		glPushMatrix();		// #0
		
		// Rotate/move the boid triangle
		glTranslatef( world->boids[i].x, world->boids[i].y, world->boids[i].z);		
		glRotatef( world->boids[i].theta_y, 0.0f, 1.0f, 0.0f ); 
	
		glPushMatrix();		// #1

		// Rotate bird about its own axis when in the process of turning left/right
		glRotatef( -5 * world->boids[i].curr_turning_y, 0.0f, 0.0f, 1.0f );

		// Totate bird about it's perp axis when in the process of turning up/down
		glRotatef( -1 * world->boids[i].curr_turning_x, 1.0f, 0.0f, 0.0f );

		//glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );
		glColor4f( 1.0f, 1.0f, 1.0f, 0.5f );
		glCallList( dl_boid );

	
		glPopMatrix();  // #1

		// Now generate the shadow!!!
		// This is currently a little inconsistent, but it gets the job done.  (sort of)
		if( use_shadows ) {
			glTranslatef( 0.0f, -world->boids[i].y, 0.0f );
			glScalef( 1.0f, 0.0f, 1.0f );
			
			// new
			glPushAttrib( GL_DEPTH_BUFFER_BIT );

			glDepthFunc(GL_ALWAYS);
			glDisable( GL_TEXTURE_2D );
			glColor4f( 0.2f, 0.2f, 0.2f, 1.0f );
			glCallList( dl_boid );

			// new
			glPopAttrib();
		}





		glPopMatrix();	// #0
	}

}

void Renderer::RenderBoidList()
{
	glNewList(dl_boid, GL_COMPILE);

	if( use_textures ) {
	  glEnable(GL_TEXTURE_2D);
	  glBindTexture( GL_TEXTURE_2D, tex_boid->id );
	}
	
  // Draw the boid
	glBegin(GL_TRIANGLES);

	//2
	glTexCoord2f( .662, .211 );
	glVertex3f( 5, 1.5, 0 );
	glTexCoord2f( .733, .243 );
	glVertex3f( 6, 1.5, 0.5 );
	glTexCoord2f( .778, .207 );
	glVertex3f( 6.5, 1.5, 0 );

	glTexCoord2f( .662, .211 );
	glVertex3f( 5, 1.5, 0 );
	glTexCoord2f( .694, .278 );
	glVertex3f( 5.3, 1.5, 1 );
	glTexCoord2f( .733, .243 );
	glVertex3f( 6, 1.5, 0.5);

	glTexCoord2f( .662, .211 );
	glVertex3f( 5, 1.5, 0 );
	glTexCoord2f( .533, .252 );
	glVertex3f( 3.5, 1.2, 0.1 );
	glTexCoord2f( .694, .278 );
	glVertex3f( 5.3, 1.5, 1 );

    ///5
	glTexCoord2f( 0.645, .325 );
	glVertex3f( 3.7, 1.5, 1.8 );
	glTexCoord2f( .533, .252 );
	glVertex3f( 3.5, 1.2, 0.1 );
	glTexCoord2f( .694, .278 );
	glVertex3f( 5.3, 1.5, 1 );

	//6
	glTexCoord2f( .645, .325 );
	glVertex3f( 3.7, 1.5, 1.8 );
	glTexCoord2f( .627, .325 );
	glVertex3f( 3.2, 1.5, 1.8 );
	glTexCoord2f( .533, .252 );
	glVertex3f( 3.5, 1.2, 0.1 );

	//7
	glTexCoord2f( .596, .345 );
	glVertex3f( 2, 1.2, 1.3 );
	glTexCoord2f( .533, .252 );
	glVertex3f( 3.5, 1.2, 0.1 );
	glTexCoord2f( .627, .325 );
	glVertex3f(  3.2, 1.5, 1.8 );

	//8
	glTexCoord2f( .596, .345 );
	glVertex3f( 2, 1.2, 1.3 );
	glTexCoord2f( .533, .252 );
	glVertex3f( 3.5, 1.2, 0.1 );
	glTexCoord2f( .462, .298 );
	glVertex3f( 2, .9, -.8 );

	glTexCoord2f( .596, .345 );
	glVertex3f( 2, 1.2, 1.3 );
	glTexCoord2f( .556, .368 );
	glVertex3f( 1, 0.8, 1 );
	glTexCoord2f( .462, .298 );
	glVertex3f( 2, .9, -.8 );
	///////////////////////////////////
	glTexCoord2f( .384, .352 );
	glVertex3f( 0.5, 0.5, -1.3 );
	glTexCoord2f( .556, .368 );
	glVertex3f( 1, 0.8, 1 );
	glTexCoord2f( .462, .298 );
	glVertex3f( 2, .9, -.8 );

	//11
	glTexCoord2f( .384, .352 );
	glVertex3f( 0.5, 0.5, -1.3 );
	glTexCoord2f( .556, .368 );
	glVertex3f( 1, 0.8, 1 );
	glTexCoord2f( .521, .384 );
	glVertex3f( 0.5, 0.6, 1 );
	

	


	//13
	glTexCoord2f( .34, .36 );
	glVertex3f( 0.5, .5, -1 );
	glTexCoord2f( .188, .329 );
	glVertex3f( 0, -.5, -3);
	glTexCoord2f( .333, .262 );
	glVertex3f( 1.7, -.5, -2.1 );

	glTexCoord2f( .286, .423 );
	glVertex3f( -0.5, .5, -1 );
	glTexCoord2f( .188, .329 );
	glVertex3f( 0, -.5, -3 );
	glTexCoord2f( .34, .36 );
	glVertex3f( 0.5, .5, -1 );

	glTexCoord2f( .196, .46 );
	glVertex3f( -1.7, -.5, -2.1 );
	glTexCoord2f( .188, .329 );
	glVertex3f( 0, -.5, -3 );
	glTexCoord2f( .286, .423 );
	glVertex3f( -0.5, .5, -1 );


	
	//16
	glTexCoord2f( 0, 0 );
	glVertex3f( 0.3, .6, 1 );
	glTexCoord2f( 1, 0 );
	glVertex3f( 0.2, .5, 1.7 );
	glTexCoord2f( 1, 1 );
	glVertex3f( 0, .5, 1.8 );

	//17
	glTexCoord2f( 0, 0 );
	glVertex3f( -0.3, .6, 1 );
	glTexCoord2f( 1, 0 );
	glVertex3f( -0.2, .5, 1.7 );
	glTexCoord2f( 1, 1 );
	glVertex3f( 0, .5, 1.8 );

	//18
	glTexCoord2f( 0, 0 );
	glVertex3f( 0.3, .6, 1 );
	glTexCoord2f( 1, 0 );
	glVertex3f( -0.3, .6, 1 );
	glTexCoord2f( 1, 1 );
	glVertex3f( 0, .5, 2);

	//20
	glTexCoord2f( .411, .415 );
	glVertex3f( 0, .7, 0);
	glTexCoord2f( .521, .384 );
	glVertex3f( 0.5, .6, 1);
	glTexCoord2f( .384, .352 );
	glVertex3f( 0.5, .5, -1 );
	////////////////////////////////////////

	glTexCoord2f( .411, .415 );
	glVertex3f( 0, .7, 0 );
	glTexCoord2f( .3, .456 );
	glVertex3f( -0.5, .5, -1.3 );
	glTexCoord2f( .384, .352 );
	glVertex3f( 0.5, .5, -1 );
	
	//22
	glTexCoord2f( .411, .415 );
	glVertex3f( 0, .7, 0);
	glTexCoord2f( .435, .488 );
	glVertex3f( 0.5, .6, 1 );
	glTexCoord2f( .521, .384 );
	glVertex3f( -0.5, .6, 1 );


	glTexCoord2f( 0.411, .415 );
	glVertex3f( 0, .7, 0 );
	glTexCoord2f( .435, .488 );
	glVertex3f( -0.5, .6, 1 );
	glTexCoord2f( .3, .456 );
	glVertex3f( -0.5, .5, -1.3 );

	//24
	glTexCoord2f( .3, .456 );
	glVertex3f( -0.5, .5, -1.3 );
	glTexCoord2f( .407, .567 );
	glVertex3f( -1, .8, 1 );
	glTexCoord2f( .435, .488 );
	glVertex3f( -0.5, .6, 1 );

	glTexCoord2f( .3, .456 );
	glVertex3f( -0.5, .5, -1.3 );
	glTexCoord2f( .407, .576 );
	glVertex3f( -1, .8, 1 );
	glTexCoord2f( .243, .607 );
	glVertex3f( -2, .9, -.8 );

	//26
	glTexCoord2f( .392, .647 );
	glVertex3f( -2, 1.2, 1.3 );
	glTexCoord2f( .407, .576 );
	glVertex3f( -1, .8, 1 );
	glTexCoord2f( .243, .607 );
	glVertex3f( -2, .9, -.8 );

	//27
	glTexCoord2f( .392, .647 );
	glVertex3f( -2, 1.2, 1.3 );
	glTexCoord2f( .235, .678 );
	glVertex3f( -3.5, 1.2, 0.1 );
	glTexCoord2f( .243, .607 );
	glVertex3f( -2, .9, -.8 );

	//28
	glTexCoord2f( .392, .647 );
	glVertex3f( -2, 1.2, 1.3 );
	glTexCoord2f( .235, .678 );
	glVertex3f( -3.5, 1.2, 0.1);
	glTexCoord2f( .37, .69 );
	glVertex3f( -3.2, 1.5, 1.8 );

	glTexCoord2f( .329, .741 );
	glVertex3f( -3.7, 1.5, 1.8 );
	glTexCoord2f( .235, .678 );
	glVertex3f( -3.5, 1.2, 0.1 );
	glTexCoord2f( .37, .69 );
	glVertex3f( -3.2, 1.5, 1.8 );
	///////////////////////////////////////////////
	glTexCoord2f( .329, .741 );
	glVertex3f( -3.7, 1.5, 1.8 );
	glTexCoord2f( .235, .678 );
	glVertex3f( -3.5, 1.2, 0.1 );
	glTexCoord2f( .282, .819 );
	glVertex3f( -5.3, 1.5, 1);

	//31
	glTexCoord2f( .2, .827 );
	glVertex3f( -5, 1.5, 0 );
	glTexCoord2f( .235, .678 );
	glVertex3f( -3.5, 1.2, 0.1 );
	glTexCoord2f( .282, .819 );
	glVertex3f( -5.3, 1.5, 1 );

	glTexCoord2f( .2, .827 );
	glVertex3f( -5, 1.5, 0 );
	glTexCoord2f( .239, .9 );
	glVertex3f( -6, 1.5, 0.5 );
	glTexCoord2f( .282, .819 );
	glVertex3f( -5.3, 1.5, 1 );
	
	//33
	glTexCoord2f(.2, .827 );
	glVertex3f( -5, 1.5, 0 );
	glTexCoord2f (.239, .9 );
	glVertex3f( -6, 1.5, 0.5 );
	glTexCoord2f (.2, .958 );
	glVertex3f( -6.5, 1.5, 0 );


	glEnd();

	glDisable( GL_TEXTURE_2D );

	glEndList();
}

/** 
 * Render some perty water
**/
void Renderer::RenderWater()
{
	glEnable(GL_TEXTURE_2D);

	static float foo = 0.0f;
	static float direction = 1.0f;
	foo += direction * 0.001f;
	if( foo >= 1.0f )
		direction = -1.0f;
	if( foo <= 0.0f )
		direction = 1.0f;

	// Draw dem waves
	glColor4f(1.0f,1.0f,1.0f,1.0f);
	glBindTexture( GL_TEXTURE_2D, tex_water1->id );

	// Wave 1
	glBegin( GL_TRIANGLE_STRIP);
	glTexCoord2f( foo, 0 );
	glVertex3f( 0.0f, 0.0f, 0.0f);
	glTexCoord2f( foo, 1 );
	glVertex3f( 0.0f, 0.0f, 1000.0f);
	glTexCoord2i( 1, 0 );
	glVertex3f( -1000.0f, 0.00f, 0.0f);
	glTexCoord2i( 1, 1 );
	glVertex3f( -1000.0f, 0.00f, 1000.0f);
	glEnd();

	// Wave 2
	glBegin( GL_TRIANGLE_STRIP);
	glTexCoord2f( 0, foo );
	glVertex3f( -500.0f, 0.0f, 0.0f);
	glTexCoord2i( 0, 1 );
	glVertex3f( -500.0f, 0.0f, -1000.0f);
	glTexCoord2f( 1, foo );
	glVertex3f( 1500.0f, 0.0f, 0.0f);
	glTexCoord2i( 1, 1 );
	glVertex3f( 1500.0f, 0.0f, -1000.0f);
	glEnd();

	// Wave 3
	glBegin( GL_TRIANGLE_STRIP);
	glTexCoord2f( 0, foo );
	glVertex3f( 980.0f, 0.0f, 0.0f);
	glTexCoord2i( 0, 1 );
	glVertex3f( 1500.0f, 0.0f, 0.0f);
	glTexCoord2f( 1, foo );
	glVertex3f( 980.0f, 0.0f, 1000.0f);
	glTexCoord2i( 1, 1 );
	glVertex3f( 1500.0f, 0.0f, 1000.0f);
	glEnd();

	// Wave 4
	glBegin( GL_TRIANGLE_STRIP);
	glTexCoord2f( 0, foo );
	glVertex3f( 1500.0f, 0.0f, 980.0f);
	glTexCoord2i( 0, 1 );
	glVertex3f( 1500.0f, 0.0f, 1500.0f);
	glTexCoord2f( 1, foo );
	glVertex3f( -500.0f, 0.0f, 980.0f);
	glTexCoord2i( 1, 1 );
	glVertex3f( -500.0f, 0.0f, 1500.0f);
	glEnd();

	glDisable(GL_TEXTURE_2D);
}

/**
 * Render Trees and such
**/
void Renderer::RenderBillboards()
{
  glEnable(GL_TEXTURE_2D);

	// Transparency blend func
	glColor4f(1.0, 1.0, 1.0, 1.0f);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glEnable(GL_BLEND);

  // Todo: Rotate towards person
	Terrain *ter = world->terrain;
	float xloc;
	float zloc; 

	for( int i = 0; i < NUM_BILLBOARDS; i++ ) {
		xloc = ter->billboards[i].x * LANDCHUNK_WIDTH;
		zloc = ter->billboards[i].z * LANDCHUNK_WIDTH;

		// two triangles per square rendered
	  glBindTexture( GL_TEXTURE_2D, tex_billboards[ ter->billboards[i].type]->id );
		glBegin(GL_TRIANGLE_STRIP);
		
		// bottom left
		glTexCoord2i( 0, 0 );
		glVertex3f( xloc, ter->billboards[i].y, zloc );

		// top left
		glTexCoord2i( 0, 1 );
		glVertex3f( xloc, ter->billboards[i].y + 20, zloc );

		// bottom right
		glTexCoord2i( 1, 0 );
		glVertex3f( xloc + 20, ter->billboards[i].y, zloc );

		// top right
		glTexCoord2i( 1, 1 );
		glVertex3f( xloc + 20, ter->billboards[i].y+20, zloc );

		glEnd();
	}

	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
}

/**
 * Set up crap
**/
void Renderer::InitGLStuff()
{
	// Thanks to lesson 7
	glShadeModel(GL_SMOOTH);						// Enable Smooth Shading
	if( use_fog )
		glClearColor(0.7f, 0.7f, 0.7f, 0.7f);		// Grey foggy background
	else
		glClearColor(0.2f, 0.2f, 1.0f, 0.7f);		// Blue sky background

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);			// Really Nice Perspective Calculations

	// OLD DEPTH STUFF
	// glEnable(GL_DEPTH_TEST);						// Enables Depth Testing
	// glDepthFunc(GL_LEQUAL);							// The Type Of Depth Testing To Do
	// glClearDepth(1.0f);							// Depth Buffer Setup

	// Special optimized depth testing
	glEnable(GL_DEPTH_TEST);						// Enables Depth Testing
	glClearDepth(1.0f);							// Depth Buffer Setup
	glClear( GL_DEPTH_BUFFER_BIT );
	glDepthRange( 0.0, 0.5 );
	glDepthFunc( GL_LESS );
	depth_flag = 1;



	// Initialize GL
  glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	glOrtho(0,1,0,1,0,100);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();


	// Is this even doing anything?
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST); // GL_NEAREST or GL_LINEAR or GL_LINEAR_MIPMAP_LINEAR
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT );


  // Init fog parameters
	glFogf( GL_FOG_START, 100.0f );
	glFogf( GL_FOG_END, 300.0f );
	glFogf( GL_FOG_DENSITY, 0.35f );
	glFogf( GL_FOG_MODE, GL_LINEAR );
	float fog_color[3] = {.7f, .7f, .7f};
	glFogfv( GL_FOG_COLOR, fog_color );

	glDisable( GL_LIGHTING );

	// Load textures
	tex_land = new Texture();
	tex_land->Generate("data/ground512.raw", 512, 512);

	tex_cloud = new Texture();
	tex_cloud->Generate("data/cloud0.raw", 256, 256);

	tex_boid = new Texture();
	tex_boid->Generate("data/good1.raw", 512, 512);//abc



	tex_water0 = new Texture();
	tex_water0->Generate("data/water0.raw", 512, 512);

	tex_water1 = new Texture();
	tex_water1->Generate("data/water1.raw", 512, 512);

	char fname[30];
	for( int i = 0; i < NUM_BILLBOARD_TYPES; i++ ) {
		tex_billboards[i] = new Texture();
		sprintf( fname, "data/tree%d.raw", i );
		tex_billboards[i]->Generate( fname, 128, 128  );
	}

	// Create some perty call lists
	dl_boid = glGenLists( 1 );
	dl_terrain = glGenLists( 1 );
	RenderTerrainList();	
	RenderBoidList();
}

void Renderer::RenderFPS()
{
	// Calculate FPS

	static int old_sec = 0;
	static int old_fps = 42;
	static int frame = 42;
	frame++;
	
	time_t timer;
	time(&timer);
	tm *clock;
	clock = localtime( &timer );
	if( clock->tm_sec != old_sec ) {
		old_sec = clock->tm_sec;
		old_fps = frame;
		frame = 0;
	}
	static char ch_fps[4];
	itoa( old_fps, ch_fps, 10 );

	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);

	glRasterPos3f( 0.0f, 0.0f, 0.0f );
	//glRasterPos4f( 0.0f, 0.0f, 0.0f, 1.0f );
	glutBitmapCharacter( GLUT_BITMAP_HELVETICA_18, 'F' );
	glutBitmapCharacter( GLUT_BITMAP_HELVETICA_18, 'P' );
	glutBitmapCharacter( GLUT_BITMAP_HELVETICA_18, 'S' );
	glutBitmapCharacter( GLUT_BITMAP_HELVETICA_18, '-' );
	for( unsigned int i = 0; i < strlen(ch_fps); i++ )
		glutBitmapCharacter( GLUT_BITMAP_HELVETICA_18, ch_fps[i] );
}

/**
 * Store the world parameter in the instance
 * so we don't need to pass it to RenderWorld every
 * frame!!!
**/
void Renderer::SetWorld(World *w)
{
  world = w;
}

/**
 * Render a single frame, using the world data
**/
void Renderer::RenderWorld()
{ 
  // Clear the background
	// This is pretty darn slow...
	// Don't clear the depth buffer... There's an optimization for this!
	//glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);			// Clear The Screen And The Depth Buffer
	//glClear( GL_DEPTH_BUFFER_BIT );
	glClear( GL_COLOR_BUFFER_BIT );

	// Depth optimization... Yay!
	if (depth_flag) {
	    glDepthFunc( GL_LESS );
	    glDepthRange( 0.0, 0.5 );
	}
	else {
	    glDepthFunc( GL_GREATER );
	    glDepthRange( 1.0, 0.5 );
	}
	depth_flag = !depth_flag;

	glLoadIdentity();

	// Rotate stuff in the opposite direction of where we're facing
	glRotatef( world->camera_ry, 0.0f, 1.0f, 0.0f ); 
	glRotatef( world->camera_rx, 1.0f * cos_lookup[ world->camera_ry ], 0.0f, 1.0f * sin_lookup[ world->camera_ry ] ); 
	
	// Move stuff in the opposite directin of where we are
	glTranslatef( -world->camera_x, -world->camera_y, -world->camera_z );

	if( use_sky )
		RenderSky();

	if( use_fog )
		glEnable(GL_FOG);
	else
		glDisable(GL_FOG);

  // Test call list
	glCallList( dl_terrain );

	RenderWater();

	if( use_billboards )
		RenderBillboards();

	RenderBoids();

	if( use_fps ) {
		glLoadIdentity();
		glTranslatef(0.4f, 0.38f, -1.0f);
		RenderFPS();
	}

	// TEST: Motion Blur!
	// This is sloooooooow
	//glAccum (GL_LOAD, 0.5f);
	//glAccum (GL_ACCUM, 0.05f);
	//glAccum (GL_MULT, 0.5f);
	//glAccum (GL_RETURN, 1.0);

  // Finish drawing, update the frame buffer
  glFinish();



	// Swap double buffer... Pretty slow as well
	if( use_doublebuffer )
	  glutSwapBuffers();
	//display();

	//delay(100);
	
}