#include "terrain.h"
#include <stdlib.h>
#include <stdio.h>

Terrain::Terrain()
{
	//use_flatland = 1;
	use_flatland = 0;
}

Terrain::~Terrain()
{

}

/**
 * Create pretty random terrain
 * TODO: Multiple chunks, fractalized???
**/
void Terrain::Generate()
{
	if( use_flatland )
		FlatGen();
	else
		RandomGen();
	
	AddBillboards();
}

/**
 * Flat 0 height terrain
**/
void Terrain::FlatGen()
{
  
	for( int ycopy = 0; ycopy < TERRAIN_SIZE; ycopy++ ) {
		for( int xcopy = 0; xcopy < TERRAIN_SIZE; xcopy++ ) {
			// Now, finally, copy to heights array
			heights[xcopy][ycopy] = 0.0f;
					
			// How much should the texture advance???
			//tex_x[ycopy][xcopy] = (float) (rand() % 20) / (float) 80 + 0.2f;
			//tex_y[ycopy][xcopy] = (float) (rand() % 20) / (float) 80 + 0.2f;
			tex_y[ycopy][xcopy] = 0.1f;
			tex_x[ycopy][xcopy] = 0.1f;
		}
	}
}

/**
 * Fill using random numbers
**/
void Terrain::RandomGen()
{
	// Temp storage... Use a grid a little bigger than the actual terrain size
	// for the smoothing calculations!!!
	const int big_size = TERRAIN_SIZE + SMOOTHNESS * 2;
	
	float h[big_size][big_size];
	float g[big_size][big_size];

	// Generate 9 hills
	int i,j;
	float sz;

  for(int n = 0; n < big_size; n++ ) {
		for(int m = 0; m < big_size; m++ ) {
			h[m][n] = 0;
			g[m][n] = 0;
		}
	}

	// Place some peaks (which then get smoothed)
  for( int hill = 0; hill < 100; hill++ ) {
		i = rand() % big_size;
		j = rand() % big_size;
		sz = 10.0f;

		h[i][j] = sz;
		g[i][j] = sz;
		//heights[i][j] = sz;
	}

	// The gourad smoothing function
  float sfunc[7][7] = { { 0.0f, 2.0f, 3.0f, 9.0f, 3.0f, 2.0f, 0.0f },
												{ 2.0f, 3.0f, 9.0f, 12.0f, 9.0f, 3.0f, 2.0f },
												{ 3.0f, 9.0f, 12.0f, 14.0f, 12.0f, 9.0f, 3.0f },
												{ 9.0f, 12.0f, 14.0f, 14.0f, 14.0f, 12.0f, 9.0f },	
												{ 3.0f, 9.0f, 12.0f, 14.0f, 12.0f, 9.0f, 3.0f },
												{ 2.0f, 3.0f, 9.0f, 12.0f, 9.0f, 3.0f, 2.0f },	
												{ 0.0f, 2.0f, 3.0f, 9.0f, 3.0f, 2.0f, 0.0f } };

	// Use a simple smothing function on the terrain we wish to return
	// Only the heights that will be put into heights[][] will be altered (to prevent out of bounds errors)
	if( SMOOTHNESS != 0 )
  for( int b = 0; b < TERRAIN_SIZE; b++ ) {
		for( int a = 0; a < TERRAIN_SIZE; a++ ) {

			// Smooth this height value
			float multiplier = 0.0f;
			for( int y = 0; y < SMOOTHNESS; y++ ) {
				for( int x = 0; x < SMOOTHNESS; x++ ) {
					g[a+SMOOTHNESS][b+SMOOTHNESS] += sfunc[x][y] * h[a-x+SMOOTHNESS][b-y+SMOOTHNESS];
					multiplier += sfunc[x][y];
				}
			}

			g[a+SMOOTHNESS][b+SMOOTHNESS] /= multiplier;

		}
	}
  
	for( int ycopy = 0; ycopy < TERRAIN_SIZE; ycopy++ ) {
		for( int xcopy = 0; xcopy < TERRAIN_SIZE; xcopy++ ) {
			// Now, finally, copy to heights array
			if( use_flatland )
				heights[xcopy][ycopy] = 0.0f;
			else
				heights[xcopy][ycopy] = g[xcopy+SMOOTHNESS][ycopy+SMOOTHNESS];
			
			// How much should the texture advance???
			//tex_x[ycopy][xcopy] = (float) (rand() % 20) / (float) 40 + 0.2f;
			
			//tex_y[ycopy][xcopy] = (float) (rand() % 20) / (float) 40 + 0.2f;
			// tex_y[ycopy][xcopy] = 1.0f;
			//tex_x[ycopy][xcopy] = 0.5f;
			//tex_y[ycopy][xcopy] = 0.5f;
			tex_x[ycopy][xcopy] = (float) (rand() % 20) / (float) 80 + 0.2f;
			tex_y[ycopy][xcopy] = (float) (rand() % 20) / (float) 80 + 0.2f;
		}
	}

}

void Terrain::AddBillboards()
{
	int posx, posz;
	for( int i = 0; i < NUM_BILLBOARDS; i++ ) {

		// Pick some random spots on the grid
		// Todo: Don't use land size in x/y/z calculation!!!  That shouyld be the renderers job.
		// Todo: floating point positions.
		// Todo: linearly interpolate the hights of spots inbetween landscape vertices
		posx = rand() % TERRAIN_SIZE;
		posz = rand() % TERRAIN_SIZE;

		billboards[i].x = (float)posx;
		billboards[i].z = (float)posz;
		// y = top of terrain
		// Todo: fix height!!!
		billboards[i].y = heights[posx][posz] * 4;
		billboards[i].type = rand() % NUM_BILLBOARD_TYPES;
	}
	

}
