/**
 * Texture.cpp
**/

#include <stdio.h>
#include <glut.h>
#include <gl\gl.h>
#include <memory.h>
#include <string.h>

#include "texture.h"

Texture::Texture()
{
	data = NULL;
	raw_filename = NULL;
}

Texture::~Texture()
{
	// Free memory
	// TODO: WHY DOES THIS CRASH!!!
	/*
	if( data )
		delete data;

	if( raw_filename )
		delete raw_filename;
	*/
}

bool Texture::LoadFile()
{
	// Try to open RAW file
	FILE *file = fopen (raw_filename, "rb");			
	if (!file) {												
		return false;
	}
	else {	

		// 24 Bit RAW Image (4 Bytes With Alpha), 256x256 Pixels
		long lCounter;										// Linear Counter

		for (int i = 0; i < height; i++)					// For Each Row
		{
			 for (int a = 0; a < width; a++)				// For Each Pixel In A Row
			 {
				// Get The RGBA Value
				lCounter=((height-1)-i) * (width*4) + (a*4);	// Flip Texture While Loading
				data[lCounter] = fgetc(file);			// Red
				data[lCounter+1] = fgetc(file);			// Green
				data[lCounter+2] = fgetc(file);			// Blue

				// Test: Black will have an alpha value of 0
				if( data[lCounter]==0 && data[lCounter+1]==0 && data[lCounter+2]==0 )
					data[lCounter+3] = 0;
				else
					data[lCounter+3] = 255;				// Alpha Value
			 }
		}
	
		// Now Close The File, We Are Done!
		fclose (file);

		// No error
		return true;
	}
	
	// Huh? We shouldn't be here!
	return false;	
}

/**
 * The main method to create a new texture.
 * This assumes that the texture data has been allocated!!!
 * @param raw_file The filename of the RAW data file
 * @param w Width of file
 * @param h Height of file
**/
bool Texture::Generate(char *raw_file, int w, int h)
{
  // Allocate memory for image (4 uchar per pixel)
	data = new unsigned char[ (w+1) * (h+1) * 4];
	if( !data )
		return false;

	raw_filename = new char[strlen(raw_file)];
	strcpy( raw_filename, raw_file );
	width = w;
	height = h;

	if( !LoadFile() ) {
		printf("ERROR LOADING FILE!!!");
		return false;
	}

	// Create Texture
	glGenTextures(1, &id);
	glBindTexture(GL_TEXTURE_2D, id);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

	gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, width, height, GL_RGBA, GL_UNSIGNED_BYTE, data);
	
	return true;
}
