/**
 * World.Cpp
**/

#include <stdlib.h>
#include "world.h"
#include "glut.h"
#include <math.h>
#include <stdio.h>

/**************** WORLD *********************/

/**
 * World Constructor
**/
World::World()
{	
	camera_x = 100;
	camera_y = 25;
	camera_z = 200;
	camera_rx = 0;
	camera_ry = 0;

	//camera_mode = CAMERA_TOP;
	//camera_mode = CAMERA_MANUAL;
	camera_mode = CAMERA_BIRD;
	//camera_mode = CAMERA_SMART;

	terrain = new Terrain();
	terrain->Generate();

  boids = new Boid[NUM_BOIDS];
	for( int i = 0; i < NUM_BOIDS; i++ ) {
		// init boid positions
		boids[i].x = (float) ( rand() % 200 + 150 );
		boids[i].y = (float) ( rand() % 40 + 10 );
		boids[i].z = (float) ( rand() % 200 + 150 );

		// Initially it doesn't want to go up/down
		boids[i].goal_y = boids[i].y;

		//boids[i].theta_z = rand() % 360;
		//boids[i].theta_x = rand() % 360;
		boids[i].theta_y = 0;
		boids[i].theta_x = 0;

		//boids[i].velocity = (float) ((1.0f/1000.0f) * (float)(rand() % 10));
		boids[i].velocity = 1.0f;

		boids[i].state = 0.5;
	}

	// Where should the boids fly to?
	mappoint_x = rand() % 1000;
	mappoint_y = rand() % 40 + 10;
	mappoint_z = rand() % 1000;

	// TEMP
	/*
	boids[0].x = 0;
	boids[0].z = 0;
	boids[0].theta_y = 10;

	boids[1].x = 0;
	boids[1].z = 10;
	boids[1].theta_y = 10;
*/

	// Init keys to off
	for( int k = 0; k < 200; k++ )
		keys[k] = 0;
}

/**
 * World Destructor
**/
World::~World()
{
	if( boids )
		delete boids;

	if( terrain )
		delete terrain;
}

/**
 * Given a x and y vector, calculate the angle formed by them
**/
int World::CalculateInverseTan( double delta_x, double delta_z )
{
	float theta;

	if( delta_x > 0 ) {
		if( delta_z >= 0 ) {
			theta = atan( delta_z / delta_x );
			theta*= 180;
			theta/= 3.1415;
			theta = 90 - theta;
		}
		else {
			theta = atan( -delta_z / delta_x );
			theta*= 180;
			theta/= 3.1415;
			theta = 90 + theta;
		}
	}
	else if( delta_x < 0 ) {
		if( delta_z >= 0 ) {
			theta = atan( delta_z / -delta_x );
			theta*= 180;
			theta/= 3.1415;
			theta = 270 + theta;
		}
		else {
			theta = atan( -delta_z / -delta_x );
			theta*= 180;
			theta/= 3.1415;
			theta = 270 - theta;
		}
	} 
	else {
		// Divide by zero check
		if( delta_z >= 0 ) {
			theta = 0;
		}
		else {
			theta = 180;
		}
	}

	// Round up/down... Todo: Is there REALLY no function for this???
	if( theta - (int)theta > 0.5 )
		return (int)theta + 1;
	else
		return (int)theta;

}

/**
 * Steer towards the center of mass
 * OTher rules too...
 * @param b The boid number to apply the rules to
**/
void World::ApplyBoidRules( int b )
{
	float cz=0, cy=0, cx=0;  // Center of mass of nearby boids
	
	double crx=0;  // Average direction vector in the x direction
	double crz=0;	// Average direction vector in the y direction 
  
	// Locate center of all boids
	for( int i = 0; i < NUM_BOIDS; i++ ) {
		// Todo: Only for nearby boids!!!
		cx += boids[i].x;
		cy += boids[i].y;
		cz += boids[i].z;
		
		crx += boids[i].velocity * sin_lookup[ boids[i].theta_y ];
		crz += boids[i].velocity * cos_lookup[ boids[i].theta_y ];
	}
	cx /= NUM_BOIDS;
	cy /= NUM_BOIDS;
	cz /= NUM_BOIDS;
	
	crz /= NUM_BOIDS;
	crx /= NUM_BOIDS;


	// Cohesion
	// Todo: Avoidance!!!

	float delta_z, delta_x;
	int com_theta;  // Center of mass theta direction for this boid

	delta_x = cx - boids[b].x;
	delta_z = cz - boids[b].z;

	boids[b].new_turning_y = 0;
	boids[b].new_turning_x = 0;

	// **** CALCULATE THE CENTER OF MASS THETA
	com_theta = CalculateInverseTan( delta_x, delta_z );

	
	
	// **** CALCULATE THE ALIGNMENT THETA
	int align_theta;
	align_theta = CalculateInverseTan( crx, crz );
	
	//printf("Crz:%f    Crx:%f  Align:%d\n", crz, crx, align_theta);
	
	//printf("dx:%f, dz:%f, curr theta:%d, com_theta:%d\n", delta_x, delta_z, boids[b].theta_y, com_theta );

	// Do cohesion turns
	boids[b].TurnTowardsY( com_theta, .15 );  //.3

	// Do alignment turns
	boids[b].TurnTowardsY( align_theta, .1 );

	// Apply a straightening force to nudge them into flying straight
	if( boids[b].curr_turning_y > 0 )
		boids[b].new_turning_y -= .05;
	else if( boids[b].curr_turning_y < 0 )
		boids[b].new_turning_y += .05;

	// Try to convince the birds to move towards a pre-picked destination
	float foo_x, foo_z;
	foo_x = mappoint_x - boids[b].x;
	foo_z = mappoint_z - boids[b].z;
	int mappoint_theta = CalculateInverseTan( foo_x, foo_z );
	boids[b].TurnTowardsY( mappoint_theta, .05 );
	

	// Now randomly choose to have this bird go up/down
	if( rand() % 100 == 42 ) {
		//boids[b].goal_y = (float)(rand() % 60) / (float)100 + 0.45;
		boids[b].goal_y = rand() % 10 + 40;
	}

	// Figure out which direction the boids should be flying (up/down)
	if( boids[b].y > boids[b].goal_y + 1.0f )
		boids[b].new_turning_x -= 0.1;
	else if( boids[b].y < boids[b].goal_y )
		boids[b].new_turning_x += 0.1;
	else if( boids[b].y < 10 * 3 )
		boids[b].new_turning_x += 0.5;
	else if( boids[b].y > 10 * 10 )
		boids[b].new_turning_x -= 0.5;
	else if( boids[b].curr_turning_x > 0 )
		boids[b].new_turning_x -= 0.1;
	else if( boids[b].curr_turning_x < 0 )
		boids[b].new_turning_x += 0.1;

	// Avoid mountains!!!
	if( boids[b].y - 5.0f < terrain->heights[(int)boids[b].x / 20][(int)boids[b].y / 20] )
		boids[b].new_turning_x += 0.5f;

}


/**
 * Move stuff in the world
 * This should happen at uniform time periods, regardless of FPS speed!
**/
void World::UpdateWorld()
{
	// Turn, then move, the birds
	
	for( int j = 0; j < NUM_BOIDS; j++ ) {
		// Mappoint check
		if( (int)boids[j].x == mappoint_x && (int)boids[j].z == mappoint_z ) {
			mappoint_x = rand() % 800 + 100;
			mappoint_z = rand() % 800 + 100;
		}

		// Figure out where this bird wants to go
		ApplyBoidRules( j );
	}

	ProcessPreBirdMoveKeys();

	for( int i = 0; i < NUM_BOIDS; i++ ) {
		// Individually move each boid
		boids[i].Move();
	}

	UpdateCamera();

}

/**
 * Handle new keypresses
 * These are handled right before a boid moves
 * These commands should be used to manipulate the birds directly
**/
void World::ProcessPreBirdMoveKeys()
{
	switch( camera_mode )	{
	case CAMERA_BIRD:
		// Simple key controls to override the bird's movements!
		if( keys[GLUT_KEY_LEFT] )
			boids[0].new_turning_y = 0.5;

		if( keys[GLUT_KEY_RIGHT] )
			boids[0].new_turning_y = -0.5;

		if( keys[GLUT_KEY_UP] ) {
			boids[0].new_turning_x = 3;
		}
			
		if( keys[GLUT_KEY_DOWN] ) {
			boids[0].new_turning_x = -3;
		}

    if( keys[GLUT_KEY_PAGE_DOWN] ) {
		  // foo
			int i = GLUT_KEY_PAGE_DOWN;
		}

		// Try to stabilize the birds vertical movement (since the bird doesn't care about X-rots yet
		if( !keys[GLUT_KEY_DOWN] && !keys[GLUT_KEY_UP] ) {

			// Boid should start stabilizing downwards
			if( boids[0].theta_x > 0 && boids[0].theta_x < 180 ) {
				boids[0].new_turning_x = -.5f;
				boids[0].curr_turning_x = boids[0].curr_turning_x < -0.5f ? -0.5f : boids[0].curr_turning_x;
			}

			// Boid should start stabilizing upwards
			if( boids[0].theta_x >= 180 && boids[0].theta_x < 360 ) {
				boids[0].new_turning_x = .5f;
				boids[0].curr_turning_x = boids[0].curr_turning_x > 0.5f ? 0.5f : boids[0].curr_turning_x;
			}

			if( boids[0].theta_x == 0 ) {
				
				if( boids[0].curr_turning_x == 0.5f ) {
					boids[0].new_turning_x = 0.0f;	
					boids[0].curr_turning_x = 0.0f;
				}
				else if( boids[0].curr_turning_x == -0.5f ) {
					boids[0].new_turning_x = 0.0f;
					boids[0].curr_turning_x = 0.0f;
				}
				if( boids[0].curr_turning_x > 0.5 )
					boids[0].new_turning_x = -0.5;
				else if( boids[0].curr_turning_x < -0.5 )
					boids[0].new_turning_x = 0.5;
				
			}

		}

		break;
	default:
		// Not much cookin here yet
		// Todo: Add more codes to make the birds do silly things!
		break;
	}

}

/**
 * Move the camera to its next position in the map
 * At this point the boids have already moved for this round
 * They should make the camera reflect the current state of the world
**/
void World::UpdateCamera()
{

	switch( camera_mode )	{
	case CAMERA_MANUAL:

		// Simple key controls
		if( keys[GLUT_KEY_LEFT] )
			camera_ry-=2;

		if( keys[GLUT_KEY_RIGHT] )
			camera_ry+=2;

		if( keys[GLUT_KEY_UP] )
			camera_rx-=2;

		if( keys[GLUT_KEY_DOWN] )
			camera_rx+=2;
		
		camera_ry += 360;
		camera_ry %= 360;
		camera_rx += 360;
		camera_rx %= 360;

		// Todo: Fix this up a little
		camera_z -= cos_lookup[camera_ry] * CAMERA_SPEED;
		camera_y -= sin_lookup[camera_rx] * CAMERA_SPEED;
		camera_x += sin_lookup[camera_ry] * CAMERA_SPEED;

		break;

	case CAMERA_TOP:
		// Look down from top
		camera_x = 100;
		camera_y = 500;
		camera_z = 150;

		camera_ry = 180;
		camera_rx = 90;

		break;

	case CAMERA_BIRD:
		// Look from directly behind bird #0
		camera_x = boids[0].x;
		camera_y = boids[0].y;
		camera_z = boids[0].z;

		camera_ry = 180 - boids[0].theta_y;
		camera_ry+=360;
		camera_ry%=360;

		// Bird moving up/down... Remain behind it!
		//camera_rx= -boids[0].theta_x;
		//camera_y -= 10 * sin_lookup[ boids[0].theta_x ];
		// Move up

		camera_x -= 13 * (sin_lookup[ boids[0].theta_y ]);
		camera_z -= 13 * (cos_lookup[ boids[0].theta_y ]);

		break;
	case CAMERA_SMART:
		
		// Follow the flock a few feet behind them
		float cz=0, cy=0, cx=0;  // Center of mass of nearby boids
		
		float crx=0;  // Average direction vector in the x direction
		float crz=0;	// Average direction vector in the y direction 
  
		// Locate center of all boids
		for( int i = 0; i < NUM_BOIDS; i++ ) {
			cx += boids[i].x;
			cy += boids[i].y;
			cz += boids[i].z;
			
			crx += boids[i].velocity * sin_lookup[ boids[i].theta_y ];
			crz += boids[i].velocity * cos_lookup[ boids[i].theta_y ];
		}
		cx /= NUM_BOIDS;
		cy /= NUM_BOIDS;
		cz /= NUM_BOIDS;
		
		crz /= NUM_BOIDS;
		crx /= NUM_BOIDS;
	
		int average_direction = CalculateInverseTan( crx, crz );
		
		// Always look at center of mass... From the viewpoint of the average_direction (wherever that may be)
		


		camera_y = cy;
		average_direction = 180 - average_direction;
		average_direction %= 360;
		camera_ry = average_direction;
		camera_rx = 0;
		camera_x = cx - 130 * crx;
		camera_z = cz - 130 * crz;
		//camera_x = cx;
		//camera_z = cz;
		
		break;
	}

}